/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "diagram_graphic.h"
#include "leveldiagram.h"
#include "geometricdiagram.h"
#include "layerdiagram.h"
#include "orderdata_tree.h"
#include "orderdata_list.h"
#include "orderdata_array.h"
#include <stdio.h>

Diagram_Graphic::Diagram_Graphic(OrderedSet *os, OrderData *d,
				 OrderedSet::data_t dummy_chains_data_type)
{
  orderedset = os;
  data = d;
  dummy_chains_datatype = dummy_chains_data_type;
  switch (dummy_chains_data_type) {
  case OrderedSet::DATA_TREE: 
    dummy_chains_data = new OrderData_Tree(); break;
  case OrderedSet::DATA_LIST: 
    dummy_chains_data = new OrderData_List(); break;
  case OrderedSet::DATA_ARRAY: 
    dummy_chains_data = new OrderData_Array();
  }
  if (!dummy_chains_data) return;
  min_x = max_x = min_y = max_y = 0;
  generationParams.generation_method = DIAGRAM_GEOMETRIC;
  generationParams.dle_fmin = true;
  generationParams.dg_fgreat = true;
  generationParams.dla_fmax = true;
  generationParams.dla_l = DLA_L_COFFMAN_GRAHAM;
  generationParams.dla_cgw = 100;
  generationParams.dla_dsdt = DLA_DD_ARRAY;
  generationParams.dla_cr = DLA_CR_ADJACENT_EXCHANGE;
}

Diagram_Graphic::~Diagram_Graphic()
{
  delete dummy_chains_data;
}

Diagram_Graphic *
Diagram_Graphic::getCopy(OrderData *new_data)
{
  Diagram_Graphic *copy;
  
  copy = new Diagram_Graphic(orderedset, new_data, dummy_chains_datatype);
  if (!copy) return NULL;
  copy->verteces = verteces;
  copy->dummy_chains_data = dummy_chains_data->getCopy();
  copy->min_x = min_x; copy->max_x = max_x; copy->min_y = min_y;
  copy->max_y = max_y;
  copy->generationParams = generationParams;
  return copy;
}

bool
Diagram_Graphic::generateDiagram(generationParams_t params)
{
  Diagram *diag = NULL;
  set<Vertex_e> verts_e;
  set<Vertex_e>::iterator iter_se;
  Vertex v(NoElement), *vert;
  int scale;
  Diagram::generationParams_t *diag_params;
  
  switch (params.generation_method) {
  case DIAGRAM_LEVEL: 
    diag = new LevelDiagram(data);
    diag_params = new LevelDiagram::generationParams_t;
    ((LevelDiagram::generationParams_t *)diag_params)->from_minimal = 
      params.dle_fmin;
    scale = 30;
    break;
  case DIAGRAM_GEOMETRIC: 
    diag = new GeometricDiagram(data);
    diag_params = new GeometricDiagram::generationParams_t;
    ((GeometricDiagram::generationParams_t *)diag_params)->from_greatest = 
      params.dg_fgreat;
    scale = 30;
    break;
  case DIAGRAM_LAYER:
    diag = new LayerDiagram(data);
    diag_params = new LayerDiagram::generationParams_t;
    ((LayerDiagram::generationParams_t *)diag_params)->from_maximal = 
      params.dla_fmax;
    switch (params.dla_l) {
    case DLA_L_LONGEST_PATH:
      ((LayerDiagram::generationParams_t *)diag_params)->layering = 
	LayerDiagram::LAYERING_LONGEST_PATH;
      break;
    case DLA_L_COFFMAN_GRAHAM:
      ((LayerDiagram::generationParams_t *)diag_params)->layering = 
	LayerDiagram::LAYERING_COFFMAN_GRAHAM;
      break;
    }
    ((LayerDiagram::generationParams_t *)diag_params)->cg_width = 
      params.dla_cgw;
    switch (params.dla_dsdt) {
    case DLA_DD_TREE:
      ((LayerDiagram::generationParams_t *)diag_params)->
	dummy_chains_data_type = LayerDiagram::DUMMY_D_TREE;
      dummy_chains_datatype = OrderedSet::DATA_TREE;
      break;
    case DLA_DD_LIST:
      ((LayerDiagram::generationParams_t *)diag_params)->
	dummy_chains_data_type = LayerDiagram::DUMMY_D_LIST;
      dummy_chains_datatype = OrderedSet::DATA_LIST;
      break;
    case DLA_DD_ARRAY:
      ((LayerDiagram::generationParams_t *)diag_params)->
	dummy_chains_data_type = LayerDiagram::DUMMY_D_ARRAY;
      dummy_chains_datatype = OrderedSet::DATA_ARRAY;
      break;
    }
    switch (params.dla_cr) {
    case DLA_CR_ADJACENT_EXCHANGE:
      ((LayerDiagram::generationParams_t *)diag_params)->crossing_reduction = 
	LayerDiagram::CROSSING_REDUCTION_ADJACENT_EXCHANGE;
      break;
    case DLA_CR_SPLIT:
      ((LayerDiagram::generationParams_t *)diag_params)->crossing_reduction = 
	LayerDiagram::CROSSING_REDUCTION_SPLIT;
      break;
    case DLA_CR_BARYCENTER:
      ((LayerDiagram::generationParams_t *)diag_params)->crossing_reduction = 
	LayerDiagram::CROSSING_REDUCTION_BARYCENTER;
      break;
    case DLA_CR_MEDIAN:
      ((LayerDiagram::generationParams_t *)diag_params)->crossing_reduction = 
	LayerDiagram::CROSSING_REDUCTION_MEDIAN;
      break;
    }
    scale = 15;
    break;
  }
  if (!diag) return false;
  diag->generateDiagram(diag_params);
  diag->getDiagram(verts_e);
  if (params.generation_method == DIAGRAM_LAYER) {
    delete dummy_chains_data;
    dummy_chains_data = ((LayerDiagram *)diag)->getDummyChainsData();
  }
  delete diag;
  verteces.clear();
  min_x = COORD_MAX; max_x = COORD_MIN; 
  min_y = COORD_MAX; max_y = COORD_MIN;
  for (iter_se = verts_e.begin(); iter_se != verts_e.end(); iter_se++) {
    if ((*iter_se).x < min_x) min_x = (*iter_se).x;
    if ((*iter_se).x > max_x) max_x = (*iter_se).x;
    if ((*iter_se).y < min_y) min_y = (*iter_se).y;
    if ((*iter_se).y > max_y) max_y = (*iter_se).y;
  };
  min_x *= scale; max_x *= scale; min_y *= scale; max_y *= scale;
  for (iter_se = verts_e.begin(); iter_se != verts_e.end(); iter_se++) {
    v = Vertex((*iter_se).element, (*iter_se).dummy, (*iter_se).x * scale,
	       (*iter_se).y * scale);
    vert = addVertex(v);
    if (vert) delete vert;
  }
  return true;
}

Vertex *
Diagram_Graphic::addVertex(Vertex v)
{
  Vertex *retval = NULL;
  set<Vertex>::iterator iter_s;
  
  iter_s = verteces.find(v);
  if (iter_s != verteces.end()) {
    retval = new Vertex(NoElement); *retval = *iter_s;
  }
  else {
    if (verteces.empty()) {
      min_x = COORD_MAX; max_x = COORD_MIN; 
      min_y = COORD_MAX; max_y = COORD_MIN;
    }
    if (v.x < min_x) min_x = v.x;
    if (v.x > max_x) max_x = v.x;
    if (v.y < min_y) min_y = v.y;
    if (v.y > max_y) max_y = v.y;
    if (v.dummy) dummy_chains_data->newElement(v.dummy);
    if (verteces.insert(v).second) {
      retval = new Vertex(NoElement); *retval = v;
    }
    orderedset->setChanged();
  }
  return retval;
}

bool
Diagram_Graphic::deleteVertex(Vertex v)
{
  set<Vertex>::iterator iter_s;
  
  iter_s = verteces.find(v);
  if (iter_s == verteces.end()) return false;
  verteces.erase(iter_s);
  if (v.dummy) dummy_chains_data->removeElement(v.dummy);
  min_x = COORD_MAX; max_x = COORD_MIN; 
  min_y = COORD_MAX; max_y = COORD_MIN;
  for (iter_s = verteces.begin(); iter_s != verteces.end(); iter_s++) {
    if ((*iter_s).x < min_x) min_x = (*iter_s).x;
    if ((*iter_s).x > max_x) max_x = (*iter_s).x;
    if ((*iter_s).y < min_y) min_y = (*iter_s).y;
    if ((*iter_s).y > max_y) max_y = (*iter_s).y;
  }
  orderedset->setChanged();
  return true; 
}

Vertex *
Diagram_Graphic::getVertex(Vertex v)
{
  Vertex *retval = NULL;
  set<Vertex>::iterator iter_s;
  
  iter_s = verteces.find(v);
  if (iter_s != verteces.end()) {
    retval = new Vertex(NoElement); *retval = *iter_s;
  }
  return retval;
}

Vertex *
Diagram_Graphic::getVertex(Element x)
{
  Vertex v(x.id);
  
  return getVertex(v);
}

void
Diagram_Graphic::clearDiagram()
{
  verteces.clear();
  dummy_chains_data->clear();
  min_x = max_x = min_y = max_y = 0;
  orderedset->setChanged();
}

Vertex *
Diagram_Graphic::getVertexInBox(Coord_t x, Coord_t y, Coord_t d)
{
  Vertex *retval = NULL;
  set<Vertex>::iterator iter_s;
  
  for (iter_s = verteces.begin(); iter_s != verteces.end(); iter_s++)
    if ((x >= (*iter_s).x - d) && (x <= (*iter_s).x + d) &&
	(y >= (*iter_s).y - d) && (y <= (*iter_s).y + d)) {
      retval = new Vertex(NoElement); *retval = *iter_s;
    }
  return retval;
}

Vertex *
Diagram_Graphic::updateVertex(Vertex v)
{
  Element *elem;
  set<Element> s;
  set<Element>::iterator iter_s;
  set<Vertex>::iterator iter_sv1, iter_sv2;
  Vertex *retval, vertex(NoElement);
  
  iter_sv1 = verteces.find(v);
  if (iter_sv1 == verteces.end()) return NULL;
  elem = getElement(v);
  if (elem) {
    retval = new Vertex(NoElement); *retval = v;
    orderedset->upperElements(s, *elem);
    for (iter_s = s.begin(); iter_s != s.end(); iter_s++) {
      vertex.element = (*iter_s).id;
      iter_sv2 = verteces.find(vertex);
      if (iter_sv2 != verteces.end()) if ((*iter_sv2).y < v.y) return retval;
    }
    s.clear();
    orderedset->lowerElements(s, *elem);
    for (iter_s = s.begin(); iter_s != s.end(); iter_s++) {
      vertex.element = (*iter_s).id;
      iter_sv2 = verteces.find(vertex);
      if (iter_sv2 != verteces.end()) if ((*iter_sv2).y > v.y) return retval;
    }
    delete elem; delete retval;
  }
  verteces.erase(iter_sv1);
  min_x = COORD_MAX; max_x = COORD_MIN; 
  min_y = COORD_MAX; max_y = COORD_MIN;
  for (iter_sv2 = verteces.begin(); iter_sv2 != verteces.end(); iter_sv2++) {
    if ((*iter_sv2).x < min_x) min_x = (*iter_sv2).x;
    if ((*iter_sv2).x > max_x) max_x = (*iter_sv2).x;
    if ((*iter_sv2).y < min_y) min_y = (*iter_sv2).y;
    if ((*iter_sv2).y > max_y) max_y = (*iter_sv2).y;
  }
  return addVertex(v);
}

Element *
Diagram_Graphic::getElement(Vertex v)
{
  Element *retval = NULL, e(v.element);
  set<Element> elems;
  set<Element>::iterator iter_s;
  
  orderedset->getElements(elems);
  iter_s = elems.find(e);
  if (iter_s != elems.end()) {
    retval = new Element(NoElement);
    *retval = *iter_s;
  }
  return retval;
}

bool
Diagram_Graphic::isEdge(Vertex vx, Vertex vy)
{
  if ((vx.element != NoElement) && (vy.element != NoElement)) {
    // vx, vy proper or mixed
    if ((vx.dummy != NoElement) && (vy.dummy != NoElement))
      // vx, vy mixed
      return (dummy_chains_data && 
	      (!dummy_chains_data->less(vx.dummy, vy.dummy, true)) &&
	      data->less(vx.element, vy.element, false));
    return data->less(vx.element, vy.element, false);
  }
  if (dummy_chains_data && (vx.dummy != NoElement) && 
      (vy.dummy != NoElement))
    // vx, vy dummy or mixed
    return dummy_chains_data->less(vx.dummy, vy.dummy, false);
  return false;
}

bool
Diagram_Graphic::addDummyEdge(Vertex vx, Vertex vy)
{
  ElementId id1, id2;

  if ((vx.dummy == NoElement) || (vy.dummy == NoElement)) return false;
  id1 = vx.dummy; id2 = vy.dummy;
  if ((vx.y > vy.y)) { id1 = vy.dummy; id2 = vx.dummy; }
  if (!dummy_chains_data->orderElements(id1, id2)) return false;
  orderedset->setChanged();
  return true;
}

bool
Diagram_Graphic::deleteDummyEdge(Vertex vx, Vertex vy)
{
  ElementId id1, id2;

  if ((vx.dummy == NoElement) || (vy.dummy == NoElement)) return false;
  id1 = vx.dummy; id2 = vy.dummy;
  if ((vx.y > vy.y)) { id1 = vy.dummy; id2 = vx.dummy; }
  if (!dummy_chains_data->removeOrder(id1, id2)) return false;
  orderedset->setChanged();
  return true;
}
