/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "geometricdiagram.h"
#include <algorithm>
#include <stdio.h>

void
GeometricDiagram::placeVertexRuleOfParallelograms(Vertex_e &vertex)
{
  /*
    For all pairs of upper elements of our new element we find their 
    common upper elements.
    Each of these upper upper elements together with the two elements
    of the pair and our new element must form a rectanle, so this gives
    us the location for our new element.
    More upper upper elements determines more locations of our new element,
    we simply choose the first one.
    If we have got no location for our new element, we place it straight
    under the upper elements in the middle.
  */
  set<ElementId> upper_elems, elems1, elems2;
  set<ElementId>::iterator iter_se;
  set<Vertex_e> upper_verts, upper_upper_verts, results;
  set<Vertex_e>::iterator iter_sv, iter_vert1, iter_vert2;
  Vertex_e vert;
  Coord_e_t sum_x = 0, sum_y = 0;
  bool other_side = false;
  
  if (generationParams->from_greatest) 
    data->upperElements(upper_elems, vertex.element);
  else data->lowerElements(upper_elems, vertex.element);
  for (iter_se = upper_elems.begin(); iter_se != upper_elems.end(); 
       iter_se++) {
    vert.element = *iter_se; vert.dummy = NoElement;
    iter_sv = verteces.find(vert);
    if (iter_sv != verteces.end()) upper_verts.insert(*iter_sv);
  }
  if (upper_verts.size() < 2) return;
  for (iter_vert1 = upper_verts.begin(), sum_y = (*iter_vert1).y;
       iter_vert1 != upper_verts.end(); iter_vert1++) {
    sum_x += (*iter_vert1).x;
    if ((generationParams->from_greatest && ((*iter_vert1).y < sum_y)) ||
	  ((!generationParams->from_greatest) && ((*iter_vert1).y > sum_y)))
      sum_y = (*iter_vert1).y;
  }
  for (iter_vert1 = upper_verts.begin(); iter_vert1 != upper_verts.end(); 
       iter_vert1++)
    for (iter_vert2 = iter_vert1, iter_vert2++; 
	 iter_vert2 != upper_verts.end(); iter_vert2++) {
      // TODO: supremum elements instead of common upper elements?
      elems1.clear(); elems2.clear();
      if (generationParams->from_greatest) {
	data->upperElements(elems1, (*iter_vert1).element);
	data->upperElements(elems2, (*iter_vert2).element);
      }
      else {
	data->lowerElements(elems1, (*iter_vert1).element);
	data->lowerElements(elems2, (*iter_vert2).element);
      }
      upper_elems.clear();
      set_intersection(elems1.begin(), elems1.end(), elems2.begin(), 
		       elems2.end(), insert_iterator<set<ElementId> >
		       (upper_elems, upper_elems.begin()));
      upper_upper_verts.clear();
      for (iter_se = upper_elems.begin(); iter_se != upper_elems.end(); 
	   iter_se++) {
	vert.element = *iter_se; vert.dummy = NoElement;
	iter_sv = verteces.find(vert);
	if (iter_sv != verteces.end()) upper_upper_verts.insert(*iter_sv);
      }
      for (iter_sv = upper_upper_verts.begin(); 
	   iter_sv != upper_upper_verts.end(); iter_sv++) {
	vert.x = (*iter_vert2).x + ((*iter_vert1).x - (*iter_sv).x);
	vert.y = (*iter_vert2).y + ((*iter_vert1).y - (*iter_sv).y);
	if (isVertexPlaceEmpty(vert.x, vert.y) &&
	    ((generationParams->from_greatest && (vert.y < sum_y)) ||
	     ((!generationParams->from_greatest) && (vert.y > sum_y))))
	  results.insert(vert);
      }
    }
  if (!results.empty()) {
    // TODO: choose other location than the first
    vertex.x = (*(results.begin())).x;
    vertex.y = (*(results.begin())).y;
  }
  else {
    vertex.x = sum_x = sum_x / (Coord_e_t)upper_verts.size();
    if (generationParams->from_greatest) vertex.y = sum_y - 2;
    else vertex.y = sum_y + 2;
    while (1) {
      if (isVertexPlaceEmpty(vertex.x, vertex.y)) break;
      if (other_side) vertex.x = sum_x + (sum_x - vertex.x);
      else vertex.x = sum_x + (sum_x - vertex.x) - 1;
      other_side = !other_side;
    }
  }
}

void
GeometricDiagram::placeVertexRuleOfLines(Vertex_e &vertex)
{
  /*
    We find all upper elements of the only upper element of our new element.
    For each of these upper upper elements we get the location of our new
    element at the extension of the line from upper upper element to upper
    element, the distance from the upper element will be the same as the
    distance of the upper element from the upper upper element.
    More upper upper elements determines more locations of our new element,
    we simply choose the first one.
    If we have got no location for our new element, we place it straight 
    under the upper element. 
  */
  set<ElementId> upper_elems;
  set<ElementId>::iterator iter_se;
  set<Vertex_e> upper_upper_verts, results;
  set<Vertex_e>::iterator iter_sv, iter_vert;
  Vertex_e vert;
  Coord_e_t x;
  bool other_side = false;

  if (generationParams->from_greatest) 
    data->upperElements(upper_elems, vertex.element);
  else data->lowerElements(upper_elems, vertex.element);
  if (upper_elems.empty()) return;
  vert.element = *(upper_elems.begin()); vert.dummy = NoElement;
  iter_vert = verteces.find(vert);
  if (iter_vert == verteces.end()) return;
  upper_elems.clear();
  if (generationParams->from_greatest) 
    data->upperElements(upper_elems, (*iter_vert).element);
  else data->lowerElements(upper_elems, (*iter_vert).element);
  for (iter_se = upper_elems.begin(); iter_se != upper_elems.end(); 
       iter_se++) {
    vert.element = *iter_se; vert.dummy = NoElement;
    iter_sv = verteces.find(vert);
    if (iter_sv != verteces.end()) upper_upper_verts.insert(*iter_sv);
  }
  for (iter_sv = upper_upper_verts.begin(); 
       iter_sv != upper_upper_verts.end(); iter_sv++) {
    vert.x = (*iter_vert).x + ((*iter_vert).x - (*iter_sv).x);
    vert.y = (*iter_vert).y + ((*iter_vert).y - (*iter_sv).y);
    if (isVertexPlaceEmpty(vert.x, vert.y)) results.insert(vert);
  }
  if (!results.empty()) { 
    //TODO: choose other location than the first
    vertex.x = (*(results.begin())).x;
    vertex.y = (*(results.begin())).y;
  }
  else {
    vertex.x = x = (*iter_vert).x;
    if (generationParams->from_greatest) vertex.y = (*iter_vert).y - 2;
    else vertex.y = (*iter_vert).y + 2;
    while (1) {
      if (isVertexPlaceEmpty(vertex.x, vertex.y)) break;
      if (other_side) vertex.x = x + (x - vertex.x);
      else vertex.x = x + (x - vertex.x) - 1;
      other_side = !other_side;
    }
  }
}

bool
GeometricDiagram::isVertexPlaceEmpty(Coord_e_t x, Coord_e_t y)
{
  set<Vertex_e>::iterator iter_sv;
  
  for (iter_sv = verteces.begin(); iter_sv != verteces.end(); iter_sv++)
    if ((x == (*iter_sv).x) && (y == (*iter_sv).y)) return false;
  return true;
}

void
GeometricDiagram::generateDiagram(Diagram::generationParams_t *params)
{
  /*
    We first place one from the maximal elements.
    Under it we place it's lower elements, ones with less lower elements 
    in the middle, ones with more lower on the edges.
    Under them go the elements which have their upper elements already
    placed in the diagram. Elements with more than 1 upper elements are
    placed under the rule of parallelograms, elements with 1 upper element
    are placed under the rule of lines.
    Previous step repeates until all elements are placed.
  */
  set<ElementId> elems1, elems2;
  set<ElementId>::iterator iter_s;
  ElementId greatest;
  GeomDiagramVertex gvert;
  set<GeomDiagramVertex> geom_diagram, gverts;
  set<GeomDiagramVertex>::iterator iter_sd;
  Vertex_e vert;
  set<Vertex_e>::iterator iter_sv;
  Coord_e_t x, y = 0;
  int elem_count = 0;
  bool other_side = false;
  
  if (!params) return;
  generationParams = (generationParams_t *)params;
  data->getElements(elems1);
  if (elems1.empty()) return;
  if (generationParams->from_greatest) data->maximalElements(elems2, elems1);
  else data->minimalElements(elems2, elems1);
  // TODO: choose another one
  greatest = *(elems2.begin());
  for (iter_s = elems1.begin(); iter_s != elems1.end(); iter_s++) {
    if (*iter_s == greatest) continue;
    gvert.element = *iter_s;
    elems2.clear();
    if (generationParams->from_greatest) data->lowerElements(elems2, *iter_s);
    else data->upperElements(elems2, *iter_s);
    gvert.lower_count = elems2.size();
    elems2.clear();
    if (generationParams->from_greatest) data->upperElements(elems2, *iter_s);
    else data->lowerElements(elems2, *iter_s);
    elems2.erase(greatest);
    gvert.upper_count = elems2.size();
    geom_diagram.insert(gvert);
  }
  vert.element = greatest; vert.dummy = NoElement;
  vert.x = 0; vert.y = y;
  if (generationParams->from_greatest) y -= 2; else y += 2;
  verteces.insert(vert);
  for (iter_sd = geom_diagram.begin(); iter_sd != geom_diagram.end(); 
       iter_sd++) {
    if ((*iter_sd).upper_count != 0) continue;
    if ((*iter_sd).lower_count > elem_count) 
      elem_count = (*iter_sd).lower_count;
    gverts.insert(*iter_sd);
  }
  x = -gverts.size() + 1;
  while (!gverts.empty()) {
    // TODO: another order of processing
    for (iter_sd = gverts.begin(); iter_sd != gverts.end(); iter_sd++)
      if ((*iter_sd).lower_count == elem_count) break;
    if (iter_sd != gverts.end()) {
      vert.element = (*iter_sd).element; vert.dummy = NoElement;
      if (other_side) { vert.x = -x; x+= 2; } 
      else vert.x = x;
      vert.y = y;
      verteces.insert(vert);
      other_side = !other_side;
      geom_diagram.erase(*iter_sd);
      gverts.erase(iter_sd);
    }
    else elem_count--;
  }
  while (!geom_diagram.empty()) {
    elem_count = 0; gverts.clear();
    for (iter_sd = geom_diagram.begin(); iter_sd != geom_diagram.end(); 
	 iter_sd++) {
      elems1.clear();
      if (generationParams->from_greatest) 
	data->upperElements(elems1, (*iter_sd).element);
      else data->lowerElements(elems1, (*iter_sd).element);
      for (iter_s = elems1.begin(); iter_s != elems1.end(); iter_s++) {
	vert.element = *iter_s; vert.dummy = NoElement;
	iter_sv = verteces.find(vert);
	if (iter_sv == verteces.end()) break;
      }
      if (iter_s != elems1.end()) continue;
      if ((*iter_sd).upper_count > elem_count) 
	elem_count = (*iter_sd).upper_count;
      gverts.insert(*iter_sd);
    }
    while (!gverts.empty()) {
      // TODO: another order of processing
      for (iter_sd = gverts.begin(); iter_sd != gverts.end(); iter_sd++)
	if ((*iter_sd).upper_count == elem_count) break;
      if (iter_sd != gverts.end()) {
	vert.element = (*iter_sd).element; vert.dummy = NoElement;
	if (elem_count > 1) placeVertexRuleOfParallelograms(vert);
	else placeVertexRuleOfLines(vert);
	verteces.insert(vert);
	geom_diagram.erase(*iter_sd);
	gverts.erase(iter_sd);
      }
      else elem_count--;
    }
  }
}
